from bs4 import BeautifulSoup as bs  # type: ignore  # install beautifulsoup4
import libcst as cst  # type: ignore
from sklearn.datasets import load_iris  # type: ignore  # install scikit-learn
import requests as r  # type: ignore
import numpy, pandas  # type: ignore  # noqa: E401

# reference imported modules for style and type checkers
bs
cst
load_iris
r
numpy
pandas
